/* todo: speedup in many places (or not, we do one fire decode every 120ms) */

//#define MAIN

#include "firecode.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#ifndef MAIN
#include "utils/log.h"
#else
#define debug(...) /*printf*/
#define error(...) /*printf*/
#endif

static int shift(int poly, int n, int r, int b)
{
  int last = (r >> (n-1)) & 1;
  r = ((r << 1) | b);
  if (last) r ^= poly;
  return r & ((1 << n) - 1);
}

static int shift_dab_data(unsigned char *data)
{
  int r = 0;
  int poly = 0x1782f;
  int n = -1; { int p = poly; while (p) { p >>= 1; n++; } }
  for (int i = 2; i <= 10; i++) {
    for (int bit = 7; bit >= 0; bit--)
      r = shift(poly, n, r, (data[i] >> bit) & 1);
  }
  for (int i = 0; i <= 1; i++) {
    for (int bit = 7; bit >= 0; bit--)
      r = shift(poly, n, r, (data[i] >> bit) & 1);
  }
  return r;
}

static int leading_zeros(int v)
{
  int r = 0;
  while ((v & 0x8000) == 0) { r++; v<<=1; }
  return r;
}

int fire_decode(unsigned char *data)
{
  int r = shift_dab_data(data);
  debug("firecode: init syndrom %16.16b\n", r);

  /* pickup first shortest error pattern */
  int error_size = 16;
  int error_pos = 0;
  int error_pattern = 0;
  for (int i = 0; i < 341; i++) {
    r = shift(0x1782f, 16, r, 0);
    int lead_zero = leading_zeros(r);
    /* get minimal error size */
    if (16 - lead_zero < error_size) {
      error_size = 16 - lead_zero;
      error_pos = i;
      error_pattern = r;
    }
    debug("firecode: bit %d: syndrom %16.16b\n", i, r);
  }
  debug("firecode: error_pos %d error_pattern %b error_size %d\n", error_pos, error_pattern, error_size);

#if 0
  /* only accept len <= 6 and pattern != 101111 */
  int pattern_length = 16 - leading_zeros(error_pattern);
  if (pattern_length > 6
      || (pattern_length == 6 && error_pattern == 0b101111))
    return -1;
#endif

  /* check that error is in the data */
  {
    /* todo: ugly, improve (was just copy/paste from below) */
    int _error_pattern = error_pattern;
    int curbit = 340 - error_pos;
    while (_error_pattern) {
      if (error_pattern & 1) {
        int byte = curbit / 8;
        if (byte > 10) return -1;
      }
      _error_pattern >>= 1;
      curbit++;
    }
  }

  /* (try to) correct */
  int curbit = 340 - error_pos;
  while (error_pattern) {
    if (error_pattern & 1) {
      int bit = curbit % 8;
      int byte = curbit / 8;
      if (byte > 10) return -1;
      /* b2b gives the byte order from the data to use in the fire decoding */
      int b2b[] = { 1, 0, 10, 9, 8, 7, 6, 5, 4, 3, 2 };
      data[b2b[byte]] ^= 1 << bit;
    }
    error_pattern >>= 1;
    curbit++;
  }

  return 0;
}

#ifdef MAIN

static int crc(unsigned char *data, int len, int poly, int crcsize, int init_value,
               int do_complement)
{
  int i;
  int j;
  int shift = init_value;
  int ret;

  poly >>= 1;

  for (i = 0; i < len; i++) {
    unsigned char d = data[i];
    for (j = 0; j < 8; j++, d<<=1) {
      int b = ((d >> 7) ^ (shift >> (crcsize-1))) & 1;
      shift = ((shift ^ (-b & poly)) << 1) | b;
    }
  }

  ret = shift & ((1 << crcsize) - 1);
  if (do_complement)
    ret = ~ret & ((1 << crcsize) - 1);

  return ret;
}

int inject_error_pattern(unsigned char *data, int pos, uint32_t error_pattern)
{
  int did_inject = 0;

  while (error_pattern) {
    if (error_pattern & 1) {
      int bit = pos % 8;
      int byte = pos / 8;
      if (byte <= 10) {
        /* b2b gives the byte order from the data to use in the fire decoding */
        int b2b[] = { 1, 0, 10, 9, 8, 7, 6, 5, 4, 3, 2 };
        data[b2b[byte]] ^= 1 << bit;
        did_inject = 1;
      }
    }
    error_pattern >>= 1;
    pos++;
  }

  return did_inject;
}

int width(int v)
{
  int first = -1;
  int last = -1;
  for (int i = 0; i < 16; i++) {
    if (v & (1 << i)) {
      first = i;
      break;
    }
  }
  if (first == -1) return 0;
  for (int i = 15; i >= 0; i--) {
    if (v & (1 << i)) {
      last = i;
      break;
    }
  }
  return last - first + 1;
}

#include <string.h>

/* check all error patterns of 1..16 bits */
void test_all(void)
{
  unsigned char data[11];
  for (int i = 0; i < 9; i++)
    data[i+2] = i;
  uint16_t v = crc(data+2, 9, 0x1782f, 16, 0, 0);
  data[0] = v >> 8;
  data[1] = v & 255;

  unsigned char fire[11];

  int max_width = 0;

  for (uint32_t error_pattern = 0; error_pattern < 65536; error_pattern++) {
    int w = width(error_pattern);
    if (w < max_width) { printf("%16.16b already done\n", error_pattern); continue; }
    if (w > max_width) max_width = w;
    int correctable = 1;
    for (int pos = 0; pos < 11*8; pos++) {
      memcpy(fire, data, 11);
      if (!inject_error_pattern(fire, pos, error_pattern)) continue;
      int v = crc(fire+2, 9, 0x1782f, 16, 0, 0);
      int v2 = (fire[0] << 8) | fire[1];
      if (v == v2) { printf("undected error!\n"); exit(1); }
      if (fire_decode(fire) !=  0 || memcmp(fire, data, 11))
        correctable = 0;
    }
    printf("%16.16b %s\n", error_pattern, correctable ? "corrected" : "uncorrectable");
  }
}

int main(void)
{
  unsigned char data[11];
  for (int i = 0; i < 9; i++) data[i+2] = i;
  uint16_t v = crc(data+2, 9, 0x1782f, 16, 0, 0);
  data[0] = v >> 8;
  data[1] = v & 255;
  /* inject error */
  data[2] ^= 0b1101101;
  //data[2] ^= 0b1000000;
  fire_decode(data);
  printf("after fire decode, crc %4.4x [%2.2x %2.2x]\n", crc(data + 2, 9, 0x1782f, 16, 0, 0), data[0], data[1]);
  test_all();
  return 0;
}

#endif /* MAIN */
