/* define MAIN for testing, then compile as:
 * gcc -o reed_solomon coding/reed_solomon.c
 */
//#define MAIN

/* I'm not sure it works with every codes out there.
 * I think we suppose alpha = 2.
 * And we expect b=0 in the decoder I think.
 * (I refer to reed-solomon-WHP031.pdf for the definitions.)
 * Those assumptions may be checked/removed if needed. For DAB+
 * Reed-Solomon coding used in ETSI TS 102 563 the code as it is is ok.
 */

#include "reed_solomon.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#ifndef MAIN
#include "utils/log.h"
#else
#define debug printf
#endif

#if 0
static uint32_t remain_1pow(int pow, uint32_t b)
{
  int i;
  int bmax;
  uint32_t ret = 0;
  for (bmax = 31; bmax > 0; bmax--) if (b & (1 << bmax)) break;
  for (i = pow; i >= 0; i--) {
    uint32_t gp1 = -((ret >> (bmax - 1)) & 1);
    uint32_t bb = i == pow;
    uint32_t n = (ret << 1) | bb;
    ret = n ^ (gp1 & b);
if (pow == 12) debug("XX %x\n", ret);
  }
  return ret;
}

static void gen_add_mul_tables(int b, int q, int *p2i, int *i2p)
{
  int a;

  /* don't be scared, p2i is passed so that p2i[-1] exists */
  p2i[-1] = 0;
  i2p[0] = -1;
  for (a = 0; a <= (1 << q) - 2; a++) {
    debug("a%d = ", a);
    int r = remain_1pow(a, b);
    p2i[a] = r;
    i2p[r] = a;
    int printed = 0;
    int first = 1;
    for (int j = q-1; j >= 0; j--) {
      if (r & (1 << j)) {
        if (!first) debug(" + ");
        if (j == 0) debug("1"); else if (j == 1) debug("a"); else debug("a%d", j);
        first = 0;
        printed = 1;
      }
    }
    if (printed == 0) debug("0");
    debug("\n");
  }
}
#endif

//static void gen_add_mul_tables(int b, int q, int *p2i, int *i2p)
static void gen_power_tables(int b, int q, int *p2i, int *i2p)
{
  int a;
  uint32_t lfsr = 1;
  uint32_t mask = 1 << (q-1);
  /* don't be scared, p2i[-1] exists */
  p2i[-1] = 0;
  i2p[0] = -1;
  for (a = 0; a <= (1 << q) - 2; a++) {
    p2i[a] = lfsr;
    i2p[lfsr] = a;
    /* shift LFSR, add b to LFSR if needed (when high bit of LFSR is 1) */
    uint32_t gp1 = (lfsr & mask) ? b : 0;
    lfsr = gp1 ^ (lfsr << 1);
  }
debug("p2i");
for (a = 0; a <= (1 << q) - 2; a++) debug(" %d", p2i[a]);
debug("\n");
debug("i2p");
for (a = 0; a <= (1 << q) - 2; a++) debug(" %d", i2p[a]);
debug("\n");
}

/* an elem a^x is stored as x, for 0 we store -1 */

int gf_add_elem(int a, int b, int *p2i, int *i2p)
{
  return i2p[p2i[a] ^ p2i[b]];
}

int gf_mul_elem(int a, int b, int n)
{
  if (a == -1 || b == -1) return -1;
  return (a + b) % n;
}

void gf_add_poly_lin(int *gf_out, int *gf_a, int *gf_b)
{
  int i;

  debug("gf_add_poly_lin a[0..100] =");
  for (i = 0; i < 100; i++) debug(" %d", gf_a[i]);
  debug("\n");
  debug("gf_add_poly_lin b[0..100] =");
  for (i = 0; i < 100; i++) debug(" %d", gf_b[i]);
  debug("\n");

  for (i = 0; i < 500; i++) gf_out[i] = gf_a[i] ^ gf_b[i];

  debug("gf_add_poly_lin res[0..100] =");
  for (i = 0; i < 100; i++) debug(" %d", gf_out[i]);
  debug("\n");
}

void gf_mul_poly_lin(int *gf_out, int *gf_a, int *gf_b, int n, int *p2i, int *i2p)
{
  int i, j;

  /* gf_out starts as pow */
  for (i = 0; i < 500; i++) gf_out[i] = -1;

  int a_order = 0;
  int b_order = 0;

  for (i = 499; i > 0; i--) if (gf_a[i]) break;
  a_order = i;
  for (i = 499; i > 0; i--) if (gf_b[i]) break;
  b_order = i;

  debug("gf_mul_poly_lin: a: ");
  for (i = 0; i <= 10+a_order; i++) debug(" %d", gf_a[i]);
  debug("\n");
  debug("gf_mul_poly_lin: b: ");
  for (i = 0; i <= 10+b_order; i++) debug(" %d", gf_b[i]);
  debug("\n");

debug("in gf_mul_poly_lin: a order %d b order %d\n", a_order, b_order);

  debug("gf_mul_poly_lin: gf_out at input: ");
  for (i = 0; i <= a_order + b_order+10; i++) debug(" %d", gf_out[i]);
  debug("\n");

  for (i = 0; i <= a_order; i++) {
    if (gf_a[i] == 0) continue;
    for (j = 0; j <= b_order; j++) {
      if (gf_b[j] == 0) continue;
      if (i + j >= 500) { debug("fatal: polys too big to mul them\n"); exit(1); }
      int mul = gf_mul_elem(i2p[gf_a[i]], i2p[gf_b[j]], n);
      gf_out[i+j] = gf_add_elem(gf_out[i+j], mul, p2i, i2p);
    }
  }

  /* convert gf_out to lin */
  for (i = 0; i < 500; i++) gf_out[i] = p2i[gf_out[i]];

  debug("gf_mul_poly_lin: gf_out at output: ");
  for (i = 0; i <= a_order + b_order+10; i++) debug(" %d", gf_out[i]);
  debug("\n");
}

void gf_mul_poly(int *gf_out, int *gf_a, int *gf_b, int t, int n, int *p2i, int *i2p)
{
  int i, j;

  for (i = 0; i < 2*t+1; i++) gf_out[i] = -1;

  for (i = 0; i < 2*t+1; i++) {
    if (gf_a[i] == -1) continue;
    for (j = 0; j < 2*t+1; j++) {
      if (gf_b[j] == -1) continue;
      if (i + j > 2*t) { debug("fatal: polys too big to mul them\n"); exit(1); }
      int mul = gf_mul_elem(gf_a[i], gf_b[j], n);
      gf_out[i+j] = gf_add_elem(gf_out[i+j], mul, p2i, i2p);
    }
  }
}

int find_pow(int n, int v1, int v2)
{
  return (v2 - v1 + n) % n;
//v2 = (v3 + v1)modxx
#if 0
  int v3;
  for (v3 = 0; v3 <= n; v3++) {
    int mul = gf_mul_elem(v1, v3, n);
    if (mul == v2) return v3;
  }
  abort();
#endif
}

void gf_div_poly(int *gf_out, int *gf_remain, int *gf_a, int *gf_b, int t, int n, int *p2i, int *i2p)
{
  int i;
  int a_order;
  int b_order;

debug("n %d\n", n);
  for (i = n; i >= 0; i--) if (gf_a[i]) break;
  a_order = i;
  if (i == -1) a_order = 1;

  for (i = n; i >= 0; i--) if (gf_b[i]) break;
  b_order = i;
  if (b_order == -1) abort();

  int q[a_order + 1];

debug("a:");
for (int j = 0; j <= a_order; j++) debug(" %d", gf_a[j]);
debug("\n");
debug("b:");
for (int j = 0; j <= b_order; j++) debug(" %d", gf_b[j]);
debug("\n");
debug("a order %d b order %d\n", a_order, b_order);
  int remain[b_order+1];

  for (i = 0; i <= b_order; i++) remain[i] = -1; //i2p[gf_a[a_order-b_order+i]];

  debug("init remain:");
  for (i = 0; i <= b_order; i++) debug(" %d", remain[i]);
  debug("\n");

  for (int j = 0; j <= a_order; j++) {
    /* shift */
    for (int l = b_order; l > 0; l--)
      remain[l] = remain[l-1];
    /* introduce */
    remain[0] = i2p[gf_a[a_order - j]];
    /* sub */
    int mul_pow;
debug("remain[%d] = %d\n", b_order, remain[b_order]);
debug("remain before addmul:");
for (int j = 0; j <= b_order; j++) debug(" %d", remain[j]);
debug("\n");
    if (remain[b_order] == -1) mul_pow = -1;
    else {
      /* g: a^v1
       * r: a^v2
       * v3 such that v3 * v1 = v2
       */
      int v1 = i2p[gf_b[b_order]];
      int v2 = remain[b_order];
      mul_pow = find_pow(n, v1, v2);
debug("v1 %d v2 %d mul_pow %d\n", v1, v2, mul_pow);
    }
    q[a_order - j] = mul_pow;
    for (int l = 0; l <= b_order; l++) {
      int v = gf_mul_elem(i2p[gf_b[l]], mul_pow, n);
debug("v[%d] %d\n", l, v);
      remain[l] = gf_add_elem(remain[l], v, p2i, i2p);
    }
debug("remain after addmul:");
for (int j = 0; j <= b_order; j++) debug(" %d", remain[j]);
debug("\n");
  }
debug("remain of div:");
for (int j = 0; j <= b_order; j++) debug(" %d", remain[j]);
debug("\n");
debug("q of div:");
for (int j = 0; j <= a_order; j++) debug(" %d", q[j]);
debug("\n");
debug("q of div (lin):");
for (int j = 0; j <= a_order; j++) debug(" %d", p2i[q[j]]);
debug("\n");

  for (int j = 0; j <= a_order; j++) gf_out[j] = p2i[q[j]];
  for (int j = 0; j <= b_order; j++) gf_remain[j] = p2i[remain[j]];

  debug("gf_out:");
  for (int j = 0; j <= n; j++) debug(" %d", gf_out[j]);
  debug("\n");
  debug("gf_remain:");
  for (int j = 0; j <= n; j++) debug(" %d", gf_remain[j]);
  debug("\n");
}

void euclid(int *a, int *b, int *p2i, int *i2p, int n, int tt, int *lambda, int *omega)  /* tt is rs->t */
{
  int q[3][500];
  int r[3][500];
  int s[3][500];
  int t[3][500];

  memset(q, 0, sizeof(q));
  memset(r, 0, sizeof(r));
  memset(s, 0, sizeof(s));
  memset(t, 0, sizeof(t));

  memcpy(r[1], a, 500*sizeof(int));
  memcpy(r[0], b, 500*sizeof(int));
  s[1][0] = 1;
  t[0][0] = 1;

  while (1) {
debug("loop\n");
    int j;
    for (j = 499; j >= 0; j--) if (r[0][j]) break;
    if (j < tt) break;

    memcpy(q[2], q[1], 500*sizeof(int));
    memcpy(q[1], q[0], 500*sizeof(int));
    memcpy(r[2], r[1], 500*sizeof(int));
    memcpy(r[1], r[0], 500*sizeof(int));
    memcpy(s[2], s[1], 500*sizeof(int));
    memcpy(s[1], s[0], 500*sizeof(int));
    memcpy(t[2], t[1], 500*sizeof(int));
    memcpy(t[1], t[0], 500*sizeof(int));

    int remain[500];
    memset(q[0], 0, sizeof(q[0]));
    memset(remain, 0, sizeof(remain));
    gf_div_poly(q[0], remain, r[2], r[1], -1, n, p2i, i2p);

//void gf_mul_poly_lin(int *gf_out, int *gf_a, int *gf_b, int n, int *p2i, int *i2p)
debug("mul 1\n");
    memset(remain, 0, sizeof(remain));
    gf_mul_poly_lin(remain, q[0], r[1], n, p2i, i2p);
    gf_add_poly_lin(r[0], r[2], remain);

debug("mul 2\n");
    memset(remain, 0, sizeof(remain));
    gf_mul_poly_lin(remain, q[0], s[1], n, p2i, i2p);
    gf_add_poly_lin(s[0], s[2], remain);

debug("mul 3\n");
    memset(remain, 0, sizeof(remain));
    gf_mul_poly_lin(remain, q[0], t[1], n, p2i, i2p);
    gf_add_poly_lin(t[0], t[2], remain);

debug("q: ");
for (int i = 0; i < 10; i++) debug("%d ", i2p[q[0][i]]);
debug(" (in i:");
for (int i = 0; i < 10; i++) debug("%d ", q[0][i]);
debug(")");
debug("\n");
debug("t: ");
for (int i = 0; i < 10; i++) debug("%d ", i2p[t[0][i]]);
debug(" (in i:");
for (int i = 0; i < 10; i++) debug("%d ", t[0][i]);
debug(")");
debug("\n");
debug("r: ");
for (int i = 0; i < 10; i++) debug("%d ", i2p[r[0][i]]);
debug(" (in i:");
for (int i = 0; i < 10; i++) debug("%d ", r[0][i]);
debug(")");
debug("\n");
  }

  memcpy(lambda, t[0], 500 * sizeof(int));
  memcpy(omega, r[0], 500 * sizeof(int));

for (int i = 0; i < 10; i++) debug("%d ", i2p[r[0][i]]);
debug("\n");
}

void reed_solomon_init(reed_solomon_t *rs, int poly, int t, int pow_start, int input_size)
{
  int i;
  int b = 0;
  int q;
  int n;

  if (!poly) { debug("bad poly, 0\n"); exit(1); }
  /* abritrary x^20 limit (or x^21, whatever) */
  if (poly > 0xfffff) { debug("bad poly, too big\n"); exit(1); }
  b = poly;

  for (i = 31; i >= 0; i--) if (b & (1 << i)) break;
  q = i;
  n = (1 << q) - 1;

  rs->p2i = malloc(sizeof(int) * (1 << q));
  if (rs->p2i == NULL) { debug("out of memory\n"); exit(1); }
  /* we need access to p2i[-1] */
  rs->p2i++;
  rs->i2p = malloc(sizeof(int) * (1 << q));
  if (rs->i2p == NULL) { debug("out of memory\n"); exit(1); }
  rs->gf = malloc(sizeof(int) * (2 * t + 1));
  if (rs->gf == NULL) { debug("out of memory\n"); exit(1); }

  gen_power_tables(b, q, rs->p2i, rs->i2p);

  /* compute (x + a^pow_start) (x + a^(pow_start+1)) ... (x + a^(pow_start+2t-1)) */
  /* in the poly, we put the power of a (-1 represents 0 ie. a^-infinity),
   * that is: a^n x^k means gf_p[k] = n
   */
  int gf_p1[2*t + 1];
  int gf_result[2*t + 1];

  for (i = 0; i < 2*t+1; i++) gf_p1[i] = gf_result[i] = -1;
  gf_result[0] = 0;

  int a;

  for (a = pow_start; a <= pow_start+2*t-1; a++) {
    gf_p1[0] = a;
    gf_p1[1] = 0;
    int gf_out[2*t + 1];
    gf_mul_poly(gf_out, gf_result, gf_p1, t, n, rs->p2i, rs->i2p);
    memcpy(gf_result, gf_out, (2*t+1) * sizeof(int));
  }

  for (i = 2*t; i >= 0; i--) {
    debug("a%d x%d + ", gf_result[i], i);
  }
  debug("\n");

  memcpy(rs->gf, gf_result, sizeof(int) * (2 * t + 1));

  rs->cw = malloc(sizeof(int) * (2 * t + 1));
  if (rs->cw == NULL) { debug("out of memory\n"); exit(1); }

  rs->n = n;
  rs->t = t;
  rs->k = rs->n - 2 * rs->t;
  rs->q = q;
  rs->input_size = input_size;

  rs->correction = malloc(t * sizeof(reed_solomon_correction_t));
  if (rs->correction == NULL) { debug("out of memory\n"); exit(1); }
}

void reed_solomon_free(reed_solomon_t *rs)
{
  free(rs->cw);
  /* reed_solomon_init() does p2i+1 */
  free(rs->p2i-1);
  free(rs->i2p);
  free(rs->gf);
  free(rs->correction);
}

/* m(x) = mk-1 x^k-1 + mk-2 x^k-2 + ... + m1 x^1 + m0 x^0
 * g(x) = g2t x^2t + g2t-1 x^2t-1 + ... + g1 x^1 + g0 x^0
 * m(x) x^n-k = mk-1 x^n-1 + mk-2 x^n-2 + ... + m1 x^n-k+1 + m0 x^n-k
 */

/* 2t = n-k */
void reed_solomon_encode(reed_solomon_t *rs, unsigned char *m, unsigned char *out)
{
  int i;
  int n = rs->n;
  int k = rs->k;
  int t = rs->t;

debug("m: ");
for (int l = k-1; l >= 0; l--) debug("a^%d x^%d + ", rs->i2p[m[l]], l);
debug("\n");

debug("m (again): ");
for (int l = rs->input_size - 1; l >= 0; l--) debug("a^%d (%2.2x) x^%d + ", rs->i2p[m[rs->input_size-1-l]], m[rs->input_size-1-l], l);
debug("\n");

  /* remainder of m * x^(n-k) divided by rs->gf */
  for (i = 0; i <= rs->t * 2; i++) rs->cw[i] = rs->i2p[m[rs->t*2-i]];

  debug("remainder:");
  for (int j = 2*t-1; j >= 0; j--)
    debug(" %d", rs->cw[j]);
  debug("\n");

  int next_m = rs->t*2+1;

debug("init cw: ");
for (int l = n-k; l >= 0; l--) debug("a^%d x^%d + ", rs->cw[l], l);
debug("\n");

  for (int j = 0; j < rs->input_size; j++) {
    int mul_pow;
    if (rs->cw[2*t] == -1) mul_pow = -1;
    else {
      /* g: a^v1
       * r: a^v2
       * v3 such that v3 * v1 = v2
       */
      int v1 = rs->gf[2*t];
      int v2 = rs->cw[2*t];
      mul_pow = find_pow(rs->n, v1, v2);
debug("v1 %d v2 %d mul_pow %d\n", v1, v2, mul_pow);
    }
    for (int l = 0; l <= n-k; l++) {
      int v = gf_mul_elem(rs->gf[l], mul_pow, rs->n);
      rs->cw[l] = gf_add_elem(rs->cw[l], v, rs->p2i, rs->i2p);
    }
debug("cw        : ");
for (int l = n-k; l >= 0; l--) debug("a^%d x^%d + ", rs->cw[l], l);
debug("\n");
    if (j == rs->input_size-1) break;
    for (int l = n-k; l > 0; l--)
      rs->cw[l] = rs->cw[l-1];
    int v;
    if (next_m < rs->input_size) v = rs->i2p[m[next_m]];
    else v = -1;
    next_m++;
    rs->cw[0] = v;
debug("cw shifted: ");
for (int l = n-k; l >= 0; l--) debug("a^%d x^%d + ", rs->cw[l], l);
debug("\n");
  }

  for (int l = n-k-1; l >= 0; l--)
    out[n-k-1-l] = rs->p2i[rs->cw[l]];

  debug("rs bytes:");
  for (int l = n-k-1; l >= 0; l--)
    debug(" %2.2x (%d)", rs->p2i[rs->cw[l]], rs->cw[l]);
  debug("\n");

  debug("rs only bytes:");
  for (int l = n-k-1; l >= 0; l--)
    debug(" %2.2x", rs->p2i[rs->cw[l]]);
  debug("\n");

  for (int l = n-k; l >= 0; l--) {
    debug("a^%d x^%d + ", rs->cw[l], l);
  }
  debug("\n");
}

/* beware! poly is in reverse */
int poly_compute(reed_solomon_t *rs, unsigned char *poly, int n, int x)
{
  x = rs->i2p[x];
  int p = x;
  int ret = rs->i2p[poly[n-1]];
  int i;

  for (i = 1; i < n; i++) {
    ret = gf_add_elem(ret, gf_mul_elem(p, rs->i2p[poly[n-1-i]], rs->n), rs->p2i, rs->i2p);
    p = gf_mul_elem(p, x, rs->n);
  }

  return rs->p2i[ret];
}

/* poly is in order, i.e. we do: poly[0] + poly[1] * x + poly[2] * x^2 + ... */
int poly_eval(reed_solomon_t *rs, int *poly, int n, int x)
{
  x = rs->i2p[x];
  int p = x;
  int ret = rs->i2p[poly[0]];
  int i;

  for (i = 1; i < n; i++) {
    ret = gf_add_elem(ret, gf_mul_elem(p, rs->i2p[poly[i]], rs->n), rs->p2i, rs->i2p);
    p = gf_mul_elem(p, x, rs->n);
  }

  return rs->p2i[ret];
}

#define BAD1 \
 0x55, 0xad, 0x7a, 0x02, 0x7f, 0x00, 0x00, 0x48, 0xd5, 0x00, 0x67, 0x4b, 0x15, 0xeb, 0x3d, 0x44, 0x82, 0xa0, 0xdc, 0x57, 0x9e, 0x8d, 0x80, 0x78, 0x4c, 0x2f, 0x74, 0x95, 0xf7, 0xbe, 0x97, 0x2c, 0x65, 0x6e, 0xbc, 0x90, 0x16, 0x00, 0x2d, 0x4e, 0x09, 0x9d, 0xdd, 0x4e, 0x00, 0x89, 0xd7, 0x89, 0x67, 0x95, 0x5b, 0x16, 0xf2, 0x99, 0xd1, 0xe8, 0x4b, 0x20, 0x6e, 0x28, 0x10, 0x87, 0x1d, 0xc3, 0x18, 0xfb, 0xf5, 0xf4, 0xda, 0x14, 0xe8, 0x29, 0x30, 0x28, 0x03, 0x34, 0x9c, 0xf5, 0x6f, 0xf1, 0xd5, 0x5b, 0x0e, 0x01, 0xa4, 0xcc, 0xdc, 0x9e, 0x56, 0xd1, 0x4d, 0x41, 0xf4, 0x09, 0x0a, 0x84, 0x52, 0x13, 0x14, 0xbd, 0x4e, 0x6e, 0x3f, 0x9b, 0x3c, 0x3f, 0x20, 0x00, 0x00, 0x01, 0xe1, 0x19, 0x35, 0xba, 0x70, 0x16, 0x75, 0xbe, 0xf6, 0x3c

#define BAD2 \
 0xf3, 0xbe, 0x55, 0xb7, 0xe9, 0x00, 0x00, 0x20, 0x65, 0x58, 0x34, 0xdc, 0x92, 0x9e, 0xbf, 0xb6, 0x38, 0xd5, 0x55, 0xc6, 0x87, 0xc6, 0xb3, 0xb8, 0x91, 0xee, 0xcd, 0x72, 0x55, 0x2d, 0xd7, 0xd2, 0xdb, 0xc1, 0x83, 0xaa, 0x16, 0x00, 0x20, 0x27, 0x19, 0x16, 0xfe, 0xc4, 0xc2, 0x36, 0x2d, 0x66, 0xea, 0x8a, 0xef, 0xc9, 0x00, 0x7f, 0xd8, 0x7f, 0x72, 0x6c, 0xf7, 0x2b, 0x39, 0x29, 0x21, 0xdb, 0x67, 0x54, 0xc3, 0xcd, 0x1f, 0x4b, 0x7d, 0xb2, 0x00, 0xff, 0x00, 0xbc, 0x4e, 0x6c, 0x25, 0x5a, 0x8c, 0x2d, 0x40, 0x18, 0x5e, 0x87, 0x77, 0xff, 0xc8, 0xb8, 0x3e, 0xdf, 0x58, 0x99, 0xfd, 0x50, 0xa4, 0xc7, 0x84, 0xc8, 0x28, 0xf0, 0x7b, 0xc4, 0x63, 0xf5, 0x00, 0x00, 0x00, 0xfc, 0x89, 0x25, 0x9f, 0xe2, 0xf9, 0xbd, 0x05, 0x27, 0x5f, 0x13

#define GOOD1 \
 0xfe, 0x37, 0xff, 0x59, 0x77, 0x00, 0x00, 0xa5, 0x96, 0xc0, 0x0d, 0x96, 0x47, 0xce, 0x4d, 0xf0, 0x09, 0x75, 0x14, 0x16, 0xba, 0x46, 0x2a, 0x12, 0x17, 0x99, 0x1b, 0x1f, 0xad, 0x73, 0x55, 0xa6, 0x78, 0xa0, 0xf0, 0x5a, 0x07, 0x00, 0x41, 0x41, 0xba, 0x24, 0x38, 0x75, 0x20, 0x0c, 0x55, 0xe7, 0x01, 0xa8, 0x19, 0xde, 0x16, 0x2a, 0x17, 0x03, 0x9d, 0xcc, 0xc9, 0x74, 0x8d, 0x7a, 0x41, 0xa5, 0xbf, 0xc6, 0x52, 0x9b, 0x19, 0xf4, 0xfe, 0xa0, 0x17, 0xfe, 0x00, 0x62, 0xd2, 0xe6, 0x75, 0xbd, 0x94, 0x03, 0x59, 0x47, 0x1f, 0xca, 0x96, 0xc8, 0xc3, 0x45, 0x44, 0x19, 0x2a, 0x08, 0xea, 0x87, 0x3c, 0x48, 0xb7, 0xeb, 0xfb, 0x3d, 0x40, 0x2f, 0x90, 0x60, 0x20, 0x00, 0x00, 0xbe, 0x5d, 0x73, 0x40, 0x6b, 0x16, 0xbf, 0xe2, 0xf5, 0x21, 0x4f

/* see doc/reed-solomon-WHP031.pdf which was used to write reed_solomon_decode */
/* http://downloads.bbc.co.uk/rd/pubs/whp/whp-pdf-files/WHP031.pdf */
/* returns number of errors corrected (may be 0) */
int reed_solomon_decode(reed_solomon_t *rs, unsigned char *data)
{
  int error_count = 0;
  /* compute syndroms, that is:
   * Si = R(a^i)
   * R is the received polynom = data[n-1] * (a^i)^(n-1) + data[n-2] * (a^i)^(n-2) + ... + data[1] * a^i + data[0]
   * except data[0] ... data[n-1] is actually reversed and shortened,
   * so let's rewrite as:
   * R is the received polynom = R[n-1] * (a^i)^(n-1) + R[n-2] * (a^i)^(n-2) + ... + R[1] * a^i + R[0]
   * and here is the relation between 'R' and 'data':
   * R[0] = data[input_size + 2*t - 1]
   * R[1] = data[input_size + 2*t - 2]
   * ...
   * R[input_size + 2*t - 1] = data[0]
   * R[input_size + 2*t] = 0
   * ...
   * R[n-1] = 0
   * then we simplify:
   * - R[x] where x >= rs->input_size + 2 * rs->t is 0, don't use it
   * - we access 'R' from high index to low, so this is 'data' from low to high,
   *   so let's use data directly
   */
  /* syndroms */
  int S[2 * rs->t];
  /* we start with a^0, etsi ts 102.563 6.1 defines the code generator polynomial starting with i = 0 */
  int a = 0;
  int n = rs->n;
  int has_error = 0;
  for (int i = 0; i < 2 * rs->t; i++) {
    /* user horner's method to compute R(a^i) */
    int v = rs->i2p[data[0]];
    for (int j = 1; j < rs->input_size + 2*rs->t; j++)
      v = gf_add_elem(gf_mul_elem(v, a, n), rs->i2p[data[j]], rs->p2i, rs->i2p);
    S[i] = rs->p2i[v];
    if (S[i]) has_error = 1;
    /* multiply by a^1, that is '1', we use pow representation */
    a = gf_mul_elem(a, 1, n);
  }
debug("S[%d]", 2 * rs->t);
for (int i = 0; i < 2 * rs->t; i++) debug(" %2.2x", S[i]);
debug("\n");

  if (has_error == 0) return 0;

  int pa[500], pb[500];
  memset(pa, 0, sizeof(pa));
  memset(pb, 0, sizeof(pb));
  pa[2 * rs->t] = 1;
  for (int i = 0; i < 2 * rs->t; i++) pb[i] = S[i];
  int lambda[500];
  int omega[500];
  euclid(pa, pb, rs->p2i, rs->i2p, n, rs->t, lambda, omega);
  int lambda_order;
  int omega_order;
  for (lambda_order = 499; lambda_order >= 0; lambda_order--) if (lambda[lambda_order]) break;
  if (lambda_order == -1) exit(1);
  for (omega_order = 499; omega_order >= 0; omega_order--) if (omega[omega_order]) break;
  if (omega_order == -1) exit(1);
debug("lambda order %d omega_order %d\n", lambda_order, omega_order);

  /* get derivate of lambda */
  int lambda_prime[500];
  int i;
  for (i = 0; i < lambda_order; i+=2) {
    lambda_prime[i] = lambda[i+1];
    lambda_prime[i+1] = 0;
  }
  lambda_prime[i] = 0; /* maybe not necessary, but we do poly_eval below, so let's do this */

debug("lambda_prime");
for (int i = 0; i < lambda_order; i++) debug(" %d", lambda_prime[i]);
debug("\n");
  for (int pos = 0; pos < rs->input_size + 2 * rs->t; pos++) {
    int x = 1 + pos + rs->n - (rs->input_size + 2 * rs->t);
    x %= n;
    int v = poly_eval(rs, lambda, lambda_order+1, rs->p2i[x]);
    if (v != 0) continue;
    /* error, let's correct */
debug("error at pos %d\n", pos);
    /* error is: Yj = Xj^(1-b) * omega(Xj^-1) / lambda_prime(Xj^-1)
     * Yj =   va     *    vb        /    vc (va, vb, vc, computed below)
     * We have b=0 for the DAB+ reed-solomon code.
     * x above is the exponent of Xj^-1
     * so we have x_inv = (n - x) % n for the inverse of x.
     */
    int x_inv = (n - x) % n;
    int va = rs->p2i[x_inv];
    int vb = poly_eval(rs, omega, omega_order+1, rs->p2i[x]);
    /* order of lambda_prime is lambda_order-1 (or -2, -1 if fine to use) */
    int vc = poly_eval(rs, lambda_prime, lambda_order, rs->p2i[x]);
    v = gf_mul_elem(rs->i2p[va], rs->i2p[vb], n);
    v = gf_mul_elem(v, (n - rs->i2p[vc]) % n, n);
debug("  error is %d\n", rs->p2i[v]);
    rs->correction[error_count].pos = pos;
    rs->correction[error_count].corrected_value = data[pos] ^ rs->p2i[v];
debug("  corrected to %d\n", rs->correction[error_count].corrected_value);
    error_count++;
    if (error_count == rs->t) {
      /* either we found all real errors or there are too many, in both cases stop here */
      debug("reed-solomon found %d errors, stops here, it's the max\n", error_count);
      break;
    }
  }
if (error_count) debug("reed-solomon: corrected %d values\n", error_count);
  return error_count;
}

#ifdef MAIN
int main(void)
{

if (1)
{
  /* with data from reed-solomon-WHP031.pdf */
  debug("test decoding\n");
  reed_solomon_t rs;
  reed_solomon_init(&rs, (1<<4) + (1<<1) + (1<<0), 2, 0, 11);
  unsigned char v[15] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  reed_solomon_encode(&rs, v, v+11);

  /* inject errors */
  v[5] = 11;
v[6] = 1;
//  v[12] = 1;
  reed_solomon_decode(&rs, v);
  return 0;
}

{
  /* with data from live recording */
  debug("test decoding\n");
  reed_solomon_t rs;
  reed_solomon_init(&rs, (1<<8) + (1<<4) + (1<<3) + (1<<2) + 1, 5, 0, 110);
  unsigned char v[] = { BAD2 };
  reed_solomon_decode(&rs, v);
  return 0;
}

#if 1
{
  reed_solomon_t rs;
  reed_solomon_init(&rs, (1<<8) + (1<<4) + (1<<3) + (1<<2) + 1, 5, 0, 110);

  unsigned char C[120] = { 0x82, 0x7b, 0x5f, 0xa1, 0x84, 0xab, 0x1b, 0x7c, 0x6b, 0xf0, 0x40, 0xf7, 0x1d, 0x19, 0x7a, 0x56, 0x88, 0x57, 0xee, 0x38, 0x1c, 0x82, 0x21, 0x29, 0x75, 0x7f, 0x9d, 0x7c, 0x4c, 0x3b, 0x82, 0x26, 0xde, 0x7b, 0xf6, 0xa8, 0x07, 0x16, 0x5e, 0x21, 0xb2, 0x77, 0xdb, 0x96, 0x49, 0xa0, 0x31, 0xfc, 0xed, 0x8a, 0xed, 0xcb, 0x2c, 0x20, 0xe3, 0x14, 0x09, 0x4c, 0x04, 0xeb, 0xf2, 0x29, 0x88, 0x02, 0x38, 0xaa, 0x00, 0xff, 0x03, 0xde, 0x4b, 0x4b, 0x4b, 0x35, 0x48, 0x92, 0x28, 0xdd, 0x95, 0x1f, 0x94, 0xe8, 0xf6, 0x4d, 0xf9, 0xbe, 0x5e, 0xff, 0x61, 0xf9, 0xc7, 0x67, 0xe2, 0x55, 0xce, 0x21, 0x48, 0x70, 0x01, 0x3c, 0x06, 0x75, 0x12, 0x16, 0xff, 0x75, 0xa5, 0xa5, 0xa5, 0x00 };

  reed_solomon_encode(&rs, C, C+110);
  debug("enc");
  for (int i = 0; i < 120; i++) debug(" %2.2x", C[i]);
  debug("\n");

  for (int i = 0; i < rs.t * 2; i++) {
    int v = poly_compute(&rs, C, 120, rs.p2i[i]);
    debug("computed %d\n", rs.i2p[v]);
    debug("----------------\n");
  }
  return 0;
}

#endif

#if 1
  /* 3.3.2 */
  reed_solomon_t rs;
  reed_solomon_init(&rs, (1<<4) + (1<<1) + 1, 3, 1, 9);

  unsigned char C[15] = { 0, 0, 0, 0, 0, 0, 0, rs.p2i[11], 0 };
  reed_solomon_encode(&rs, C, C+9);

  debug("out:");
  for (int i = 0; i < 15; i++) debug(" %d", C[i]);
  debug("\n");

  debug("out (power):");
  for (int i = 0; i < 15; i++) debug(" %d", rs.i2p[C[i]]);
  debug("\n");

  /* inject errors */
  C[6] = rs.p2i[0];
  C[12] = rs.p2i[3];

  debug("out+error (power):");
  for (int i = 0; i < 15; i++) debug(" %d", rs.i2p[C[i]]);
  debug("\n");

C[109]++;
  for (int i = 0; i < rs.t * 2; i++) {
    int v = poly_compute(&rs, C, 15, rs.p2i[i+1]);
    debug("computed %d\n", rs.i2p[v]);
    debug("----------------\n");
  }

  //int lambda[4];
  //lambda[0] = 1;

//void gf_div_poly(int *gf_out, int *gf_a, int *gf_b, int t, int n, int *p2i, int *i2p)
int gf_out[500] = {0};
int gf_remain[500] = {0};
int gf_a[500] = {0};
int gf_b[500] = {0};
#if 0
gf_a[4] = 1;
gf_a[1] = 1;
gf_a[0] = 1;
gf_b[2] = 1;
gf_b[1] = 1;
gf_b[0] = 1;
#endif
#if 0
gf_a[3] = 1;
gf_b[1] = 1;
#endif
#if 1
gf_a[6] = 3;
gf_a[0] = 2;

gf_b[1] = 1;
gf_b[0] = 1;
#endif
debug("div:\n");
gf_div_poly(gf_out, gf_out, gf_a, gf_b, rs.t, rs.n, rs.p2i, rs.i2p);

{
int a = 0;
int b = 0;
int c = gf_add_elem(a, b, rs.p2i, rs.i2p);
debug("a %d b %d a+b %d\n", a, b, gf_add_elem(a, b, rs.p2i, rs.i2p));
debug("a %d c %d a+c %d\n", a, c, gf_add_elem(a, c, rs.p2i, rs.i2p));
debug("b %d c %d b+c %d\n", b, c, gf_add_elem(b, c, rs.p2i, rs.i2p));
}

/* 4.3.1.2 */
  reed_solomon_init(&rs, (1<<4) + (1<<1) + 1, 3, 1, 15);
memset(gf_a, 0, sizeof(gf_a));
memset(gf_b, 0, sizeof(gf_b));
gf_a[6] = 1;

gf_b[5] = rs.p2i[10];
gf_b[4] = 0;
gf_b[3] = 1;
gf_b[2] = rs.p2i[5];
gf_b[1] = 1;
gf_b[0] = 1;

  gf_div_poly(gf_out, gf_remain, gf_a, gf_b, rs.t, rs.n, rs.p2i, rs.i2p);

  reed_solomon_init(&rs, (1<<4) + (1<<1) + 1, 3, 1, 15);
memset(gf_a, 0, sizeof(gf_a));
memset(gf_b, 0, sizeof(gf_b));
gf_a[6] = rs.p2i[0];
gf_b[0] = rs.p2i[13];
gf_b[1] = rs.p2i[4];
gf_b[2] = rs.p2i[8];
gf_b[3] = rs.p2i[2];
gf_b[4] = rs.p2i[3];
gf_b[5] = rs.p2i[8];
debug("-------------------\n");
debug("going euclid\n");
int lambda[500];
int omega[500];
euclid(gf_a, gf_b, rs.p2i, rs.i2p, rs.n, rs.t, lambda, omega);

  reed_solomon_init(&rs, (1<<4) + (1<<1) + 1, 3, 1, 15);
memset(gf_a, 0, sizeof(gf_a));
memset(gf_b, 0, sizeof(gf_b));
gf_a[6] = rs.p2i[0];
gf_b[0] = rs.p2i[0];
gf_b[1] = rs.p2i[0];
gf_b[2] = rs.p2i[5];
gf_b[3] = rs.p2i[0];
gf_b[4] = rs.p2i[-1];
gf_b[5] = rs.p2i[10];
debug("-------------------\n");
debug("going euclid\n");
euclid(gf_a, gf_b, rs.p2i, rs.i2p, rs.n, rs.t, lambda, omega);

  return 0;
#endif
}
#endif
