#ifndef _REED_SOLOMON_H_
#define _REED_SOLOMON_H_

typedef struct {
  int pos;
  int corrected_value;
} reed_solomon_correction_t;

typedef struct {
  int *gf;       /* generator polynomial */
  int *p2i;
  int *i2p;
  int *cw;
  int n, k, t, q;
  int input_size;
  /* 'correction' is of size 't' and populated by reed_solomon_decode() */
  reed_solomon_correction_t *correction;
} reed_solomon_t;

void reed_solomon_init(reed_solomon_t *rs, int poly, int t, int pow_start,
                       int input_size);
void reed_solomon_free(reed_solomon_t *rs);

void reed_solomon_encode(reed_solomon_t *rs, unsigned char *in, unsigned char *out);
/* reed_solomon_decode() returns the number of errors corrected and
 * stores the corrections in rs->correction
 */
int reed_solomon_decode(reed_solomon_t *rs, unsigned char *in);

#endif /* _REED_SOLOMON_H_ */
