#include "reed_solomon.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#if 0
static uint32_t remain_1pow(int pow, uint32_t b)
{
  int i;
  int bmax;
  uint32_t ret = 0;
  for (bmax = 31; bmax > 0; bmax--) if (b & (1 << bmax)) break;
  for (i = pow; i >= 0; i--) {
    uint32_t gp1 = -((ret >> (bmax - 1)) & 1);
    uint32_t bb = i == pow;
    uint32_t n = (ret << 1) | bb;
    ret = n ^ (gp1 & b);
if (pow == 12) printf("XX %x\n", ret);
  } 
  return ret;
}
  
static void gen_add_mul_tables(int b, int q, int *p2i, int *i2p)
{ 
  int a;

  /* don't be scared, p2i is passed so that p2i[-1] exists */
  p2i[-1] = 0;
  i2p[0] = -1;
  for (a = 0; a <= (1 << q) - 2; a++) {
    printf("a%d = ", a);
    int r = remain_1pow(a, b);
    p2i[a] = r;
    i2p[r] = a;
    int printed = 0;
    int first = 1;
    for (int j = q-1; j >= 0; j--) {
      if (r & (1 << j)) {
        if (!first) printf(" + ");
        if (j == 0) printf("1"); else if (j == 1) printf("a"); else printf("a%d", j);
        first = 0;
        printed = 1;
      }
    } 
    if (printed == 0) printf("0");
    printf("\n");
  }
} 
#endif

//static void gen_add_mul_tables(int b, int q, int *p2i, int *i2p)
static void gen_power_tables(int b, int q, int *p2i, int *i2p)
{
  int a;
  uint32_t lfsr = 1;
  /* don't be scared, p2i[-1] exists */
  p2i[-1] = 0;
  i2p[0] = -1;
  for (a = 0; a <= (1 << q) - 2; a++) {
    p2i[a] = lfsr;
    i2p[lfsr] = a;
    int printed = 0;
    int first = 1;
    printf("a%d = ", a);
    for (int j = q-1; j >= 0; j--) {
      if (lfsr & (1 << j)) {
        if (!first) printf(" + ");
        if (j == 0) printf("1"); else if (j == 1) printf("a"); else printf("a%d", j);
        first = 0;
        printed = 1;
      }
    } 
    if (printed == 0) printf("0");
    printf(" lfsr %x b %x\n", lfsr, b);
    /* rotate LFSR, add b to LFSR if needed (when high bit of LFSR is 1) */
    uint32_t gp1 = (lfsr >> (q - 1)) & 1 ? b : 0;
    lfsr = gp1 ^ (lfsr << 1);
  } 
}

/* an elem a^x is stored as x, for 0 we store -1 */

int gf_add_elem(int a, int b, int *p2i, int *i2p)
{   
  int sum = i2p[p2i[a] ^ p2i[b]];
  return sum;
}   
    
int gf_mul_elem(int a, int b, int q)
{
 int mul;
  if (a == -1 || b == -1) mul = 0; else mul = (a + b) % ((1 << q) - 1) + 1;
  return mul - 1;
} 

void gf_mul_poly(int *gf_out, int *gf_a, int *gf_b, int t, int q, int *p2i, int *i2p)
{
  int i, j;

  for (i = 0; i < 2*t+1; i++) gf_out[i] = -1;
  
  for (i = 0; i < 2*t+1; i++) {
    if (gf_a[i] == -1) continue;
    for (j = 0; j < 2*t+1; j++) {
      if (gf_b[j] == -1) continue;
      if (i + j > 2*t) { printf("fatal: polys too big to mul them\n"); exit(1); }
      int mul = gf_mul_elem(gf_a[i], gf_b[j], q);
      gf_out[i+j] = gf_add_elem(gf_out[i+j], mul, p2i, i2p);
    } 
  } 
}

void reed_solomon_init(reed_solomon_t *rs, int poly, int t, int pow_start, int _)
{
  int i;
  int b = 0;
  int q;

  if (!poly) { printf("bad poly, 0\n"); exit(1); }
  /* abritrary x^20 limit (or x^21, whatever) */
  if (poly > 0xfffff) { printf("bad poly, too big\n"); exit(1); }
  b = poly;

  for (i = 31; i >= 0; i--) if (b & (1 << i)) break;
  q = i;

  rs->p2i = malloc(sizeof(int) * (1 << q));
  if (rs->p2i == NULL) { printf("out of memory\n"); exit(1); }
  /* we need access to p2i[-1] */
  rs->p2i++;
  rs->i2p = malloc(sizeof(int) * (1 << q));
  if (rs->i2p == NULL) { printf("out of memory\n"); exit(1); }
  rs->gf = malloc(sizeof(int) * (2 * t + 1));
  if (rs->gf == NULL) { printf("out of memory\n"); exit(1); }

  gen_power_tables(b, q, rs->p2i, rs->i2p);

  /* compute (x + a^pow_start) (x + a^(pow_start+1)) ... (x + a^(pow_start+2t-1)) */
  /* in the poly, we put the power of a (-1 represents 0 ie. a^-infinity),
   * that is: a^n x^k means gf_p[k] = n
   */
  int gf_p1[2*t + 1];
  int gf_result[2*t + 1];

  for (i = 0; i < 2*t+1; i++) gf_p1[i] = gf_result[i] = -1;
  gf_result[0] = 0;

  int a;

  for (a = pow_start; a <= pow_start+2*t-1; a++) {
    gf_p1[0] = a;
    gf_p1[1] = 0;
    int gf_out[2*t + 1];
    gf_mul_poly(gf_out, gf_result, gf_p1, t, q, rs->p2i, rs->i2p);
    memcpy(gf_result, gf_out, (2*t+1) * sizeof(int));
  }

  for (i = 2*t; i >= 0; i--) {
    printf("a%d x%d + ", gf_result[i], i);
  }
  printf("\n");

  memcpy(rs->gf, gf_result, sizeof(int) * (2 * t + 1));

  rs->cw = malloc(sizeof(int) * (2 * t + 1));
  if (rs->cw == NULL) { printf("out of memory\n"); exit(1); }

  rs->n = (1 << q) - 1;
  rs->t = t;
  rs->k = rs->n - 2 * rs->t;
  rs->q = q;
}

int find_pow(int q, int v1, int v2)
{
  int v3;
  for (v3 = 0; v3 < 1<<q; v3++) {
    int mul = gf_mul_elem(v1, v3, q);
    if (mul == v2) return v3;
  }
  abort();
}

/* m(x) = mk-1 x^k-1 + mk-2 x^k-2 + ... + m1 x^1 + m0 x^0
 * g(x) = g2t x^2t + g2t-1 x^2t-1 + ... + g1 x^1 + g0 x^0
 * m(x) x^n-k = mk-1 x^n-1 + mk-2 x^n-2 + ... + m1 x^n-k+1 + m0 x^n-k
 */

/* 2t = n-k */
void reed_solomon_encode(reed_solomon_t *rs, unsigned char *m)
{
  int i;
  int n = rs->n;
  int k = rs->k;
  int t = rs->t;

printf("m: ");
for (int l = k-1; l >= 0; l--) printf("a^%d x^%d + ", rs->i2p[m[l]], l);
printf("\n");

printf("m (again): ");
for (int l = k-1; l >= 0; l--) printf("a^%d (%2.2x) x^%d + ", rs->i2p[m[l]], m[l], l);
printf("\n");

  /* remainder of m * x^(n-k) divided by rs->gf */
  for (i = 0; i < rs->t * 2; i++) rs->cw[i] = rs->i2p[m[i + k - t*2]];

  printf("remainder:");
  for (int j = 2*t-1; j >= 0; j--)
    printf(" %d", rs->cw[j]);
  printf("\n");

  int next_m = k-1;
  for (int j = 2 * t; j >= 0; j--) {
    int v;
    if (next_m >= 0) v = rs->i2p[m[next_m]];
    else v = -1;
    next_m--;
    rs->cw[j] = v;
  }
printf("init cw: ");
for (int l = n-k; l >= 0; l--) printf("a^%d x^%d + ", rs->cw[l], l);
printf("\n");

  for (int j = 0; j < k; j++) {
    int mul_pow;
    if (rs->cw[2*t] == -1) mul_pow = -1;
    else {
      /* g: a^v1
       * r: a^v2
       * v3 such that v3 * v1 = v2
       */
      int v1 = rs->gf[2*t];
      int v2 = rs->cw[2*t];
      mul_pow = find_pow(rs->q, v1, v2);
printf("v1 %d v2 %d mul_pow %d\n", v1, v2, mul_pow);
    }
    for (int l = 0; l <= n-k; l++) {
      int v = gf_mul_elem(rs->gf[l], mul_pow, rs->q);
      rs->cw[l] = gf_add_elem(rs->cw[l], v, rs->p2i, rs->i2p);
    }
printf("cw        : ");
for (int l = n-k; l >= 0; l--) printf("a^%d x^%d + ", rs->cw[l], l);
printf("\n");
    if (j == k-1) break;
    for (int l = n-k; l > 0; l--)
      rs->cw[l] = rs->cw[l-1];
    int v;
    if (next_m >= 0) v = rs->i2p[m[next_m]];
    else v = -1;
    next_m--;
    rs->cw[0] = v;
printf("cw shifted: ");
for (int l = n-k; l >= 0; l--) printf("a^%d x^%d + ", rs->cw[l], l);
printf("\n");
  }

  printf("rs bytes:");
  for (int l = n-k-1; l >= 0; l--)
    printf(" %2.2x (%d)", rs->p2i[rs->cw[l]], rs->cw[l]);
  printf("\n");

  printf("rs only bytes:");
  for (int l = n-k-1; l >= 0; l--)
    printf(" %2.2x", rs->p2i[rs->cw[l]]);
  printf("\n");

  for (int l = n-k; l >= 0; l--) {
    printf("a^%d x^%d + ", rs->cw[l], l);
  }
  printf("\n");
}

int main(void)
{
  reed_solomon_t rs;
  //reed_solomon_init(&rs, (1<<8) + (1<<4) + (1<<3) + (1<<2) + 1, 5, 0);
  reed_solomon_init(&rs, (1<<4) + (1<<1) + 1, 3, 1, 0);

  unsigned char m[] = {
    0,
    rs.p2i[11], 0, 0, 0, 0, 0, 0, 0
  };

  reed_solomon_encode(&rs, m);

  printf("____________\n");

  unsigned char m2[] = {
    0, 0, 0,
    rs.p2i[5], 0, 0, 0, 0, 0
  };

  reed_solomon_encode(&rs, m2);

  printf("____________\n");

  unsigned char m3[] = {
    5, 2, 1, 6, 8, 3, 10, 15, 4
  };

  reed_solomon_encode(&rs, m3);

  printf("____________\n");

  unsigned char C[] = { 0x82, 0x7b, 0x5f, 0xa1, 0x84, 0xab, 0x1b, 0x7c, 0x6b, 0xf0, 0x40, 0xf7, 0x1d, 0x19, 0x7a, 0x56, 0x88, 0x57, 0xee, 0x38, 0x1c, 0x82, 0x21, 0x29, 0x75, 0x7f, 0x9d, 0x7c, 0x4c, 0x3b, 0x82, 0x26, 0xde, 0x7b, 0xf6, 0xa8, 0x07, 0x16, 0x5e, 0x21, 0xb2, 0x77, 0xdb, 0x96, 0x49, 0xa0, 0x31, 0xfc, 0xed, 0x8a, 0xed, 0xcb, 0x2c, 0x20, 0xe3, 0x14, 0x09, 0x4c, 0x04, 0xeb, 0xf2, 0x29, 0x88, 0x02, 0x38, 0xaa, 0x00, 0xff, 0x03, 0xde, 0x4b, 0x4b, 0x4b, 0x35, 0x48, 0x92, 0x28, 0xdd, 0x95, 0x1f, 0x94, 0xe8, 0xf6, 0x4d, 0xf9, 0xbe, 0x5e, 0xff, 0x61, 0xf9, 0xc7, 0x67, 0xe2, 0x55, 0xce, 0x21, 0x48, 0x70, 0x01, 0x3c, 0x06, 0x75, 0x12, 0x16, 0xff, 0x75, 0xa5, 0xa5, 0xa5, 0x00 };
  unsigned char C2[110];
  int i;
  for (i = 0; i < 110; i++) C2[i] = C[109-i];

  unsigned char M[245];
  memset(M, 0, sizeof(M));
  for (i = 0; i < 110; i++) M[135+i] = C[i];

  reed_solomon_init(&rs, (1<<8) + (1<<4) + (1<<3) + (1<<2) + 1, 5, 0, 0);
  reed_solomon_encode(&rs, M);

  printf("____________\n");

  for (i = 0; i < 110; i++) M[135+i] = C2[i];
  reed_solomon_encode(&rs, M);

  printf("____________\n");
  memset(M, 0, sizeof(M));
  for (i = 0; i < 110; i++) M[i] = C[i];
  reed_solomon_encode(&rs, M);

  printf("____________\n");
  memset(M, 0, sizeof(M));
  for (i = 0; i < 110; i++) M[i] = C2[i];
  reed_solomon_encode(&rs, M);

  printf("____________\n");
  unsigned char D[] = { 0x95, 0x89, 0x91, 0x84, 0x9e, 0x00, 0xc1, 0x23, 0x88, 0xd0, 0x1c, 0x58, 0xc5, 0x22, 0xee, 0xcc, 0x7c, 0x4e, 0x72, 0x62, 0xc3, 0xb2, 0xb2, 0xd7, 0xc8, 0x4e, 0xbc, 0x63, 0xc8, 0xf9, 0x26, 0x7f, 0xd7, 0xd0, 0x06, 0x96, 0xf6, 0x8d, 0x5c, 0x20, 0x8b, 0xb1, 0xda, 0x2d, 0xb4, 0x48, 0x7c, 0xfc, 0x03, 0x0c, 0x5f, 0x6e, 0xf0, 0x09, 0x1d, 0xa5, 0xcf, 0x10, 0xec, 0xc2, 0x02, 0x88, 0x33, 0x7d, 0x35, 0x64, 0x0c, 0x6f, 0x02, 0x4b, 0x4b, 0x4b, 0x29, 0x48, 0x13, 0x1a, 0x61, 0xeb, 0x4d, 0x17, 0x99, 0x0b, 0x43, 0x02, 0x10, 0xeb, 0x50, 0x51, 0xc7, 0x33, 0x59, 0xbd, 0x24, 0x68, 0x19, 0x0c, 0x22, 0x1b, 0x8f, 0x2c, 0x8d, 0xe0, 0x63, 0xdc, 0x1f, 0xa5, 0xa5, 0xa5, 0x00, 0xc0 };
  memset(M, 0, sizeof(M));
  for (i = 0; i < 110; i++) M[i] = D[109-i];
  reed_solomon_encode(&rs, M);

  return 0;
}
