/* this tool transforms 1 bit IQ data (what they call "hardbits") from
 * https://github.com/williamyang98/DAB-Radio/releases
 * to time domain fully framed IQ data that dabs can read
 */

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <math.h>

#include "phy/signal.h"
#include "coding/frequency_interleaving.h"
#include "utils/fft.h"

int read_frame(uint8_t *out)
{
  return fread(out, 230400/8, 1, stdin) == 1;
}

void output(complex float *b, int sample_count)
{
  uint8_t bout[sample_count * 4];

  for (int i = 0; i < sample_count; i++) {
    int16_t _i = creal(b[i]) * 32000;
    int16_t _q = cimag(b[i]) * 32000;
    /* a bit low, put more gain */
    _i *= 10;
    _q *= 10;
    bout[i * 4] = _i & 255;
    bout[i * 4 + 1] = (_i >> 8) & 255;
    bout[i * 4 + 2] = _q & 255;
    bout[i * 4 + 3] = (_q >> 8) & 255;
  }

  fwrite(bout, 4, sample_count, stdout);
}

void decompress(uint8_t *in, complex float *out)
{
  uint8_t p[K*2];
  for (int i = 0; i < K*2; i++) {
    int byte = i / 8;
    int bit = i % 8;
    p[i] = (in[byte] >> bit) & 1;
  }
  for (int i = 0; i < K; i++)
    out[i] = 1/sqrt(2) * (1 - 2 * p[i] + I * (1 - 2 * p[i+K]));
}

void interleave(frequency_interleaver *fl, complex float *in, complex float *out)
{
  /* out is [-768..-1 0 1..768], that is K+1 freq */
  /* DC is null */
  out[768] = 0;
  /* positive frequencies [1..768] */
  for (int j = 0; j < K/2; j++)
    out[769 + j] = in[fl->k2n[FK(j+1)]];
  /* negative frequencies [-768..-1] */
  for (int j = 0; j < K/2; j++)
    out[j] = in[fl->k2n[FK(j-K/2)]];
}

int main(void)
{
  uint8_t frame[230400/8];
  complex float silence[T_NULL] = {0};
  complex float first_symbol[T_S];
  complex float symbol[T_S];
  complex float *next;
  complex float *prev;
  int frame_count = 0;
  fft_data *fft = init_fft(2048);

  complex float qpsk[K];
  complex float symbol1[K+1];
  complex float symbol2[K+1];
  complex float phase_ref[K+1] = { 0 };
  compute_phase_ref(phase_ref);

  do_signal(first_symbol, phase_ref);
  //do_signal_fast(fft, first_symbol, phase_ref);

  frequency_interleaver *fl = new_frequency_interleaver();

  while (1) {
    if (!read_frame(frame)) break;
    frame_count++;
    output(silence, T_NULL);
    output(first_symbol, T_S);
    memcpy(symbol1, phase_ref, (K+1)*sizeof(complex float));
    prev = symbol1;
    next = symbol2;
    for (int i = 0; i < 75; i++) {
      fprintf(stderr, "doing frame %d symbol %d\n", frame_count-1, i);
      decompress(frame + i * K*2/8, qpsk);
      interleave(fl, qpsk, next);
      /* dqpsk */
      for (int j = 0; j <= K; j++)
        next[j] *= prev[j];
      do_signal_fast(fft, symbol, next);
      output(symbol, T_S);
      complex float *x = prev; prev = next; next = x;
    }
  }
  free_fft(fft);
  fprintf(stderr, "%d frames\n", frame_count);
  fflush(stdout);
  return 0;
}
