#include "gui.h"

#include <stdlib.h>
#include <pthread.h>

#include "utils/log.h"
#include "gui/gui.h"
#include "view/view.h"

typedef struct {
  gui *gui;
  view *iq[76];
} dab_gui;

static void create_widgets(dab_gui *g, int size)
{
  widget *main_window;
  widget *top_container;
  int i, j;

  main_window = new_toplevel_window(g->gui, 8 * 120, 8 * 120, "dabs");
  top_container = new_container(g->gui, VERTICAL);
  widget_add_child(g->gui, main_window, top_container, -1);

  for (i = 0; i < 9; i++) {
    widget *line = new_container(g->gui, HORIZONTAL);
    widget_add_child(g->gui, top_container, line, -1);
    for (j = 0; (i < 8 && j < 9) || (i == 8 && j < 4); j++) {
      char sym[64];
      sprintf(sym, "sym %d", i * 9 + j + 1);
      widget *iq = new_xy_plot(g->gui, 60, 60, sym, 40);
      int scale = i == 0 && j == 0 ? 40 : 2000;
      xy_plot_set_range(g->gui, iq, -scale, scale, -scale, scale);
      widget_add_child(g->gui, line, iq, -1);
      g->iq[i * 9 + j] = new_view_xy(size, 10, g->gui, iq, new_color(g->gui, "#500"), XY_FORCED_MODE);
    }
  }
}

void *gui_thread(void *_g)
{
  dab_gui *g = _g;
  gui_loop(g->gui);
  return NULL;
}

void *start_gui(int size)
{
  dab_gui *g;
  g = calloc(1, sizeof(dab_gui)); if (g == NULL) abort();
  g->gui = gui_init();

  create_widgets(g, size);

  pthread_t t;
  if (pthread_create(&t, NULL, gui_thread, g)) {
    error("pthread_create failed\n");
    exit(1);
  }

  return g;
}

void plot_iq(void *_g, int plot, float complex *qpsk, int size)
{
  dab_gui *g = _g;
  float x[size];
  float y[size];
  int i;

  if (plot < 1 || plot > 77) exit(1);

  for (i = 0; i < size; i++) {
    x[i] = creal(qpsk[i]);
    y[i] = cimag(qpsk[i]);
  }
  g->iq[plot - 1]->append(g->iq[plot - 1], x, y, size);
}
